;;; 
;;; Runtime include file:
;;; Contains the minimal set of binding necessary
;;; for running a fully expanded program.
;;;

(define ex:unspecified (if #f #f))
(define ex:library (make-rtd 'library '#((immutable name) (immutable envs) (immutable exports) (immutable imports) (immutable builds) (immutable visiter) (immutable invoker) (immutable build) (mutable visited?) (mutable invoked?))))

(define ex:make-library 
   (let ((constructor (rtd-constructor ex:library)))
	(lambda (name envs exports imports builds visiter invoker build)
	   (constructor name envs exports imports builds visiter invoker build #f #f))))

;(define (ex:make-library name envs exports imports builds visiter invoker build)
;  (list name envs exports imports builds visiter invoker build))

(define ex:library-name (rtd-accessor ex:library 'name))
(define ex:library-envs (rtd-accessor ex:library 'envs))
(define ex:library-exports (rtd-accessor ex:library 'exports))
(define ex:library-imports (rtd-accessor ex:library 'imports))
(define ex:library-builds  (rtd-accessor ex:library 'builds))
(define ex:library-visiter (rtd-accessor ex:library 'visiter))
(define ex:library-invoker (rtd-accessor ex:library 'invoker))
(define ex:library-build   (rtd-accessor ex:library 'build))
(define ex:library-visited? (rtd-accessor ex:library 'visited?))
(define ex:library-invoked? (rtd-accessor ex:library 'invoked?))

(define ex:library-visited?-set! (rtd-mutator ex:library 'visited?))
(define ex:library-invoked?-set! (rtd-mutator ex:library 'invoked?))

(define ex:imported '())
(define (ex:import-libraries-for imports builds phase importer run-or-expand)
  (define (import-libraries imports builds phase)
    (for-each (lambda (import build)
                (let ((name   (car import))
                      (levels (cdr import)))
                  (for-each (lambda (level)
                              (import-library name build (+ phase level)))
                            levels)))
              imports
              builds)
    (values))
  (define (import-library name build phase)
    (if (not (member (cons name (cons phase run-or-expand)) ex:imported))
        (let ((library (ex:lookup-library name)))
          (or (not build)
              (eq? build (ex:library-build library))
              (assertion-violation 'import "Client was expanded against a different build of this library" name))
          (import-libraries (ex:library-imports library) 
                            (ex:library-builds library)
                            phase)
          (importer library phase ex:imported)
          (set! ex:imported (cons (cons name (cons phase run-or-expand)) ex:imported)))))
  (import-libraries imports builds phase))

(define (ex:import-libraries-for-run imports builds phase)
  (ex:import-libraries-for imports 
                           builds
                           phase 
                           (lambda (library phase imported)
                             (if (and (= phase 0)
                                      (not (ex:library-invoked? library)))
                                 (begin 
                                   ((ex:library-invoker library))
                                   (ex:library-invoked?-set! library #t))))
                           'run))



(define ex:register-library! #f)
(define ex:lookup-library    #f)
(define ex:register-catalogs! #f)
(define ex:catalog #f)
(let ((table '())
      (catalog (make-hash-table)))

     (define (ex:load-library name)
	(let ((version (find-version catalog name (lambda rest #t) '(gambit-binary gambit))))
		(cond 
		    ((and version (version-property version 'dynlib)) (load (version-uri version)))
		    ((version? version) (ex:run-r6rs-sequence (ex:read-library version)))
		    (else (assertion-violation 'lookup-library "Library not registered in catalog!" name (hash-table->alist catalog))))
		(ex:lookup-library name)))


      #;(let ((library (hash-table-ref/default catalog name #f)))
       (if library
	(let* ((versions (library-versions library))
		
	       ;first try to find the gambit-binary and load that
	       (gambit-binary (call/cc (lambda (return)
					(for-each 
					 (lambda (version) (if (memq 'gambit-binary (version-features version)) (return version)))
					 versions) #f)))
		(source (or gambit-binary (ex:read-library (call/cc (lambda (return)
								  (for-each 
								   (lambda (version) 
								    (if (not (memq 'gambit-binary (version-features version))) (return version)))
								   versions)))))))

	(if gambit-binary 
	     (load (version-uri gambit-binary))
	     (ex:run-r6rs-sequence (list source))) ;no gambit binary; load first source file
	  (ex:lookup-library name))
	(assertion-violation 'lookup-library "Library not registered in catalog!" name (hash-table->alist catalog))))      
			

     (define version-reference->predicate
      (let ()
       (define (and? v->p references) 
	(let ((predicates (map v->p references)))
	 (lambda (v) 
	  (call/cc (lambda (return)
		    (for-each (lambda (p) (if (not (p v)) (return #f))) predicates)
		    (return #t))))))

       (define (or? v->p references)
	(let ((predicates (map v->p references)))
	 (lambda (v)
	  (call/cc (lambda (return)
		    (for-each (lambda (p) (if (p v) (return #t))) predicates)
		    (return #f))))))

       (define (not? v->p reference)
	(let ((predicate (v->p reference)))
	 (lambda (v) (not (predicate v)))))

       (define (subversion-predicate subversion)
	(if (integer? subversion) (lambda (sv) (eqv? subversion sv))
	 (case (car subversion)
	  ((>=) (lambda (sv) (>= sv (cadr subversion))))
	  ((<=) (lambda (sv) (<= sv (cadr subversion))))
	  ((and) (and? subversion-predicate (cdr subversion)))
	  ((or) (or? subversion-predicate (cdr subversion)))
	  ((not) (not? subversion-predicate (cadr subversion))))))

	(define (compare? version predicates)
	 (call/cc (lambda (return)
		   (for-each (lambda (p sv) 
			      (if p (return #t)
			       (if (not (p sv)) (return #f))) predicates version))
		   (return #t))))

	(define (make-list k fill) (if (zero? k) '() (cons fill (make-list (- k 1) fill))))

	(lambda (version-reference)
	 (if (null? version-reference) (lambda (v) #t)
	  (case (car version-reference)
	   ((and) (and? version-predicate (cdr version-reference)))
	   ((or) (or? version-predicate (cdr version-reference)))
	   ((not) (not? version-predicate (cadr version-reference)))
	   (else (let* ((sv-predicates (map subversion-predicate version-reference))
			(sv-len (length sv-predicates)))
		  (lambda (v)
		   (cond 
		    ((= sv-len (length v)) (compare? v sv-predicates)) 
		    ((< sv-len (length v)) (compare? v (append sv-predicates (make-list (- (length v) sv-len) #t))))
		    (else #f))))))))))

  (set! ex:register-catalogs! (lambda (catalogs)
	(for-each (lambda (library) (hash-table-set! catalog (library-name library) library)) 
		  (apply append (map catalog->libraries catalogs)))
  (set! ex:catalog (lambda () (pp (hash-table->alist catalog))))
))

  (set! ex:register-library! 
        (lambda (library)
          (set! table (cons (cons (ex:library-name library) library) table))
	  (set! ex:imported (filter (lambda (entry)
                                      (not (equal? (ex:library-name library) 
                                                   (car entry))))
                                    ex:imported))))

  (set! ex:lookup-library 
        (lambda (name)
          (let ((library (assoc name table)))
            (if library
                (cdr library)
		(begin
		   (ex:load-library name)
		   ;(ex:lookup-library name)
))))))


;; Only instantiate part of the bootstrap library 
;; that would be needed for invocation at runtime.

(ex:register-library! 
 (let ((error (lambda () 
                (assertion-violation 
                 'runtime.scm
                 "Attempt to use runtime instance of (core primitive-macros) for expansion.  Make sure expander.scm is loaded after runtime.scm."))))
   (ex:make-library
    '(core primitive-macros)
    ;; envs
    error
    ;; exports
    '()
    ;; imported-libraries
    '()
    ;; builds
    '()
    ;; visiter
    error
    ;; invoker
    (lambda () (values))
    ;; build
    'system)))
